/*
 * smtp.c
 * This file is part of LibSMTPc
 *
 * Copyright (C) 2022 - Félix Arreola Rodríguez
 *
 * LibSMTPc is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * LibSMTPc is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with LibSMTPc; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, 
 * Boston, MA  02110-1301  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

#include <errno.h>
#include <ctype.h>

#include "smtp.h"
#include "base64.h"

#define SMTP_WANT_FLAGS_OR_POLL(var, flags, poll) \
	do { \
	if ((var & flags) == 0) { \
		poll = flags; \
		return SMTP_POLL; \
	} \
	} while (0)

enum {
	SMTP_LINE_INVALID = 0,
	SMTP_LINE_CODE,
	SMTP_LINE_CONTINUE
};

enum {
	SMTP_INTERNAL_FREE = 0,
	SMTP_INTERNAL_WRITE,
	SMTP_INTERNAL_READ,
};

struct _SMTPSession {
	/* El socket de datos */
	int sock;
	
	/* Estado de la sessión en general */
	int estado;
	
	/* Estado interno, lectura, escritura */
	int internal;
	
	/* Variable que controla el poller */
	int poll;
	
	/* La conexión con el servidor */
	struct sockaddr_storage server;
	
	/* El buffer de lectura de la red */
	char read_buffer[8192];
	size_t read_count;
	int ceof;
	
	/* El último código de 3 digitos del último comando enviado */
	int last_code;
	
	/* El ehlo que vamos a utilizar */
	char helo_name[300];
	
	/* Utilizado para saber si sigo buscando lineas de continuación de un comando multilineas */
	char cont_code[6];
	char *user;
	char *passwd;
	char last_cmd[8192];
	char last_message[1024];
	
	/* El stream que usaremos para enviar el correo */
	FILE *data_stream;
	
	/* Función para llamar cuando lea entre lineas */
	SMTPLineFunc line_func;
	void *line_func_data;
};

int _smtp_session_internal_connect_cont (SMTPSession *session);
int _smtp_session_internal_banner_code (SMTPSession *session);
int _smtp_session_internal_send_helo (SMTPSession *session);
int _smtp_session_internal_helo_code (SMTPSession *session);
int _smtp_session_internal_start_auth_login (SMTPSession *session);
int _smtp_session_internal_auth_login_code (SMTPSession *session);
int _smtp_session_internal_send_user (SMTPSession *session);
int _smtp_session_internal_user_code (SMTPSession *session);
int _smtp_session_internal_send_pass (SMTPSession *session);
int _smtp_session_internal_pass_code (SMTPSession *session);
int _smtp_session_internal_send_cmd (SMTPSession *session);
int _smtp_session_internal_cmd_code (SMTPSession *session);
int _smtp_session_internal_send_close (SMTPSession *session);
int _smtp_session_internal_close_code (SMTPSession *session);
int _smtp_session_internal_send_data (SMTPSession *session);
int _smtp_session_internal_data_code (SMTPSession *session);
int _smtp_session_internal_data_write_finish (SMTPSession *session);
int _smtp_session_internal_data_write (SMTPSession *session);

int smtp_session_read_bytes (SMTPSession *session);
int smtp_session_readline (SMTPSession *session, char *buffer, int max);
int _smtp_session_validate_response_line (SMTPSession *session, char *buffer, int size);

SMTPSession *smtp_session_create (void) {
	SMTPSession *smtp;
	
	smtp = (SMTPSession *) malloc (sizeof (SMTPSession));
	
	if (smtp == NULL) return NULL;
	
	memset (smtp, 0, sizeof (SMTPSession));
	
	smtp->sock = -1;
	
	return smtp;
}

void smtp_session_free (SMTPSession *session) {
	smtp_session_close_and_clean (session);
	
	free (session);
}

void smtp_session_close_and_clean (SMTPSession *session) {
	if (session->user != NULL) {
		free (session->user);
		session->user = NULL;
	}
	
	if (session->passwd != NULL) {
		free (session->passwd);
		session->passwd = NULL;
	}
	
	if (session->sock >= 0) {
		close (session->sock);
		session->sock = -1;
	}
	
	session->internal = SMTP_INTERNAL_FREE;
	session->estado = SMTP_STATE_CLOSED;
}

int smtp_session_connect (SMTPSession *session, struct sockaddr *server, socklen_t addr_len, const char *helo_name) {
	int c;
	int res;
	
	session->poll = SMTP_WANT_NONE;
	
	c = socket (server->sa_family, SOCK_STREAM, 0);
	
	if (c < 0) {
		return SMTP_ERROR;
	}
	
	fcntl (c, F_SETFL, O_NONBLOCK);
	
	session->estado = SMTP_STATE_TRYING_CONNECT;
	memcpy (&session->server, server, addr_len);
	
	strncpy (session->helo_name, helo_name, sizeof (session->helo_name));
	
	res = connect (c, server, addr_len);
	
	if (res < 0) {
		if (errno == EINPROGRESS) {
			session->sock = c;
			session->poll = SMTP_WANT_POLL_OUT;
			session->internal = SMTP_INTERNAL_WRITE;
			session->estado = SMTP_STATE_TRYING_CONNECT;
			return SMTP_POLL;
		} else {
			session->sock = -1;
			close (c);
			
			return SMTP_ERROR;
		}
	} else {
		/* Conexión exitosa */
		session->sock = c;
		session->internal = SMTP_INTERNAL_READ;
		session->estado = SMTP_STATE_BANNER;
		session->read_count = 0;
		return smtp_session_continue (session, SMTP_WANT_POLL_IN);
	}
}

int smtp_session_continue (SMTPSession *session, int what) {
	char linea[8192];
	size_t linea_size;
	int res;
	int e;
	
	if (session->internal == SMTP_INTERNAL_FREE) {
		return SMTP_ERROR;
	}
	
	if (session->internal == SMTP_INTERNAL_WRITE) {
		SMTP_WANT_FLAGS_OR_POLL (what, SMTP_WANT_POLL_OUT, session->poll);
		
		/* Se logró el evento esperado */
		switch (session->estado) {
			case SMTP_STATE_TRYING_CONNECT:
				return _smtp_session_internal_connect_cont (session);
			case SMTP_STATE_HELO:
				return _smtp_session_internal_send_helo (session);
			case SMTP_STATE_AUTH_LOGIN:
				return _smtp_session_internal_start_auth_login (session);
			case SMTP_STATE_AUTH_LOGIN_USER:
				return _smtp_session_internal_send_user (session);
			case SMTP_STATE_AUTH_LOGIN_PASS:
				return _smtp_session_internal_send_pass (session);
			case SMTP_STATE_CMD:
				return _smtp_session_internal_send_cmd (session);
			case SMTP_STATE_CLOSING:
				return _smtp_session_internal_send_close (session);
			case SMTP_STATE_DATA:
				return _smtp_session_internal_send_data (session);
			case SMTP_STATE_SENDING_DATA:
				return _smtp_session_internal_data_write (session);
		}
	} else if (session->internal == SMTP_INTERNAL_READ) {
		SMTP_WANT_FLAGS_OR_POLL (what, SMTP_WANT_POLL_IN, session->poll);
		
		/* Cuando sea una lectura, esperamos una linea con un código de respuesta, el que sea */

smtp_continue_read_line_again:
		/* Leer una linea */
		linea_size = smtp_session_readline (session, linea, sizeof (linea));
		if (linea_size == SMTP_POLL) {
			/* Si al intentar leer, no hay linea, revisar si podemos leer algunos bytes */
			res = smtp_session_read_bytes (session);
			if (res == SMTP_ERROR) {
				smtp_session_close_and_clean (session);
				return SMTP_ERROR;
			} else if (res == SMTP_POLL) {
				/* No hay bytes, hacer polling */
				session->poll = SMTP_WANT_POLL_IN;
				return SMTP_POLL;
			} else if (session->ceof == 1) {
				/* Cierre de la conexión, limpiar y cerrar */
				if (session->estado == SMTP_STATE_CLOSING) {
					smtp_session_close_and_clean (session);
					return SMTP_SUCCESS;
				} else {
					smtp_session_close_and_clean (session);
					return SMTP_ERROR;
				}
			}
			/* Con los bytes restantes de la lectura previa, intentar leer la linea */
			linea_size = smtp_session_readline (session, linea, sizeof (linea));
		}
		
		/* Procesar la linea */
		res = _smtp_session_validate_response_line (session, linea, linea_size);
		
		if (res == SMTP_LINE_INVALID) {
			/* Esperaba una respuesta numérica */
			smtp_session_close_and_clean (session);
			return SMTP_ERROR;
		} else if (res == SMTP_LINE_CONTINUE) {
			/* Llamarnos recursivamente para procesar la siguiente linea */
			goto smtp_continue_read_line_again;
		}
		/* Linea buena, guardar el estado y el mensaje de la linea */
		/* Estamos conectados */
		strncpy (session->last_message, &linea[4], sizeof (session->last_message) - 1);
		e = strlen (session->last_message) - 1;
		while (e > 0 && (session->last_message[e] == '\r' || session->last_message[e] == '\n')) {
			session->last_message[e] = 0;
			e--;
		}
		
		session->last_code = ((linea[0] - '0') * 100) + ((linea[1] - '0') * 10) + (linea[2] - '0');
		/* Correr la función correcta para pasar al siguiente estado */
		switch (session->estado) {
			case SMTP_STATE_BANNER:
				return _smtp_session_internal_banner_code (session);
			case SMTP_STATE_HELO:
				return _smtp_session_internal_helo_code (session);
			case SMTP_STATE_AUTH_LOGIN:
				return _smtp_session_internal_auth_login_code (session);
			case SMTP_STATE_AUTH_LOGIN_USER:
				return _smtp_session_internal_user_code (session);
			case SMTP_STATE_AUTH_LOGIN_PASS:
				return _smtp_session_internal_pass_code (session);
			case SMTP_STATE_CMD:
				return _smtp_session_internal_cmd_code (session);
			case SMTP_STATE_CLOSING:
				return _smtp_session_internal_close_code (session);
			case SMTP_STATE_DATA:
				return _smtp_session_internal_data_code (session);
			case SMTP_STATE_SENDING_DATA:
				return _smtp_session_internal_data_write_finish (session);
		}
	}
}

int _smtp_session_validate_response_line (SMTPSession *session, char *buffer, int size) {
	if (size < 4 && session->cont_code[0] == 0) {
		return SMTP_LINE_INVALID;
	}
	
	if (session->cont_code[0] != 0) {
		/* Estamos leyendo una continuación, si son digitos más espacio, es la linea final */
		if (size >= 4 && strncmp (session->cont_code, buffer, 3) == 0 && buffer[3] == ' ') {
			session->cont_code[0] = 0; /* Ya no es continuación */
			return SMTP_LINE_CODE;
		} else {
			/* Lo que sea la linea, es continuación de lo anterior */
			
			/* Si el usuario instaló una función de entre lineas, llamarla */
			if (session->line_func != NULL) {
				session->line_func (session, buffer, session->line_func_data);
			}
			return SMTP_LINE_CONTINUE;
		}
	} else {
		if (!isdigit (buffer[0]) || !isdigit (buffer[1]) || !isdigit (buffer[2])) {
			return SMTP_LINE_INVALID;
		}
		
		if (buffer[3] == '-') {
			/* Empieza una continuación multi-linea */
			session->cont_code[0] = buffer[0];
			session->cont_code[1] = buffer[1];
			session->cont_code[2] = buffer[2];
			session->cont_code[3] = 0;
			
			/* Si el usuario instaló una función de entre lineas, llamarla */
			if (session->line_func != NULL) {
				session->line_func (session, buffer, session->line_func_data);
			}
			
			return SMTP_LINE_CONTINUE;
		} else {
			return SMTP_LINE_CODE;
		}
	}
}

int _smtp_session_internal_connect_cont (SMTPSession *session) {
	int optval;
	socklen_t optlen;
	int res;
	
	optval = -1;
	
	optlen = sizeof (optval);
	
	res = getsockopt (session->sock, SOL_SOCKET, SO_ERROR, &optval, &optlen);

	if (res < 0) {
		smtp_session_close_and_clean (session);

		return SMTP_ERROR;
	}
	
	if (optval != 0) {
		smtp_session_close_and_clean (session);

		return SMTP_ERROR;
	}
	
	/* Pasamos a la lectura del banner */
	session->internal = SMTP_INTERNAL_READ;
	session->estado = SMTP_STATE_BANNER;
	session->read_count = 0;
	return smtp_session_continue (session, SMTP_WANT_POLL_IN);
}

int _smtp_session_internal_banner_code (SMTPSession *session) {
	if (session->last_code >= 200 && session->last_code < 300) {
		session->internal = SMTP_INTERNAL_WRITE;
		session->estado = SMTP_STATE_HELO;
		return smtp_session_continue (session, SMTP_WANT_POLL_OUT);
	} else {
		smtp_session_close_and_clean (session);
		return SMTP_ERROR;
	}
}

int _smtp_session_internal_send_helo (SMTPSession *session) {
	char linea[1024];
	int res, linea_size;
	
	linea_size = snprintf (linea, sizeof (linea), "EHLO %s\r\n", session->helo_name);
	res = write (session->sock, linea, linea_size);
	
	if (res < 0 && errno == EAGAIN) {
		session->poll = SMTP_WANT_POLL_OUT;
		session->internal = SMTP_INTERNAL_WRITE;
		return SMTP_POLL;
	} else if (res < linea_size) {
		/* Enviamos menos de lo que deberíamos o error, TODO: revisar este caso */
		smtp_session_close_and_clean (session);
		
		return SMTP_ERROR;
	}
	/* Envio completo */
	session->internal = SMTP_INTERNAL_READ;
	
	return smtp_session_continue (session, SMTP_WANT_POLL_IN);
}

int _smtp_session_internal_helo_code (SMTPSession *session) {
	if (session->last_code >= 200 && session->last_code < 300) {
		session->internal = SMTP_INTERNAL_FREE;
		session->estado = SMTP_STATE_READY;
		return SMTP_SUCCESS;
	} else {
		smtp_session_close_and_clean (session);
		return SMTP_ERROR;
	}
}

int _smtp_session_internal_start_auth_login (SMTPSession *session) {
	int res, linea_size;
	char *auth = "AUTH LOGIN\r\n";
	
	linea_size = strlen (auth);
	res = write (session->sock, auth, linea_size);
	
	if (res < 0 && errno == EAGAIN) {
		session->poll = SMTP_WANT_POLL_OUT;
		session->internal = SMTP_INTERNAL_WRITE;
		return SMTP_POLL;
	} else if (res < linea_size) {
		smtp_session_close_and_clean (session);
		
		return SMTP_ERROR;
	}
	
	/* Empezamos el AUTH LOGIN */
	session->internal = SMTP_INTERNAL_READ;
	
	return smtp_session_continue (session, SMTP_WANT_POLL_IN);
}

int _smtp_session_internal_auth_login_code (SMTPSession *session) {
	if (session->last_code >= 300 && session->last_code < 400) {
		session->estado = SMTP_STATE_AUTH_LOGIN_USER;
		session->internal = SMTP_INTERNAL_WRITE;
		return smtp_session_continue (session, SMTP_WANT_POLL_OUT);
	} else if (session->last_code >= 500) {
		session->estado = SMTP_STATE_READY;
		session->internal = SMTP_INTERNAL_FREE;
		return SMTP_ERROR;
	} else {
		smtp_session_close_and_clean (session);
		return SMTP_ERROR;
	}
}

int _smtp_session_internal_send_user (SMTPSession *session) {
	char linea[1024];
	int res, linea_size;
	size_t b64_len;
	char *b64;
	
	b64 = base64_encode (session->user, strlen (session->user), &b64_len);
	linea_size = snprintf (linea, sizeof (linea), "%s\r\n", b64);
	free (b64);
	res = write (session->sock, linea, linea_size);
	
	if (res < 0 && errno == EAGAIN) {
		session->poll = SMTP_WANT_POLL_OUT;
		session->internal = SMTP_INTERNAL_WRITE;
		return SMTP_POLL;
	} else if (res < linea_size) {
		/* Enviamos menos de lo que deberíamos o error, TODO: revisar este caso */
		smtp_session_close_and_clean (session);
		
		return SMTP_ERROR;
	}
	/* Envio completo */
	session->internal = SMTP_INTERNAL_READ;
	
	return smtp_session_continue (session, SMTP_WANT_POLL_IN);
}

int _smtp_session_internal_user_code (SMTPSession *session) {
	if (session->last_code >= 300 && session->last_code < 400) {
		session->estado = SMTP_STATE_AUTH_LOGIN_PASS;
		session->internal = SMTP_INTERNAL_WRITE;
		return smtp_session_continue (session, SMTP_WANT_POLL_OUT);
	} else if (session->last_code >= 200 && session->last_code < 300) {
		session->internal = SMTP_INTERNAL_FREE;
		session->estado = SMTP_STATE_READY;
		
		/* FIXME: activar la bandera de logueado */
		if (session->passwd) {
			free (session->passwd);
			session->passwd = NULL;
		}
		return SMTP_SUCCESS;
	} else {
		smtp_session_close_and_clean (session);
		return SMTP_ERROR;
	}
}

int _smtp_session_internal_send_pass (SMTPSession *session) {
	char linea[1024];
	int res, linea_size;
	size_t b64_len;
	char *b64;
	
	b64 = base64_encode (session->passwd, strlen (session->passwd), &b64_len);
	linea_size = snprintf (linea, sizeof (linea), "%s\r\n", b64);
	free (b64);
	res = write (session->sock, linea, linea_size);
	
	if (res < 0 && errno == EAGAIN) {
		session->poll = SMTP_WANT_POLL_OUT;
		session->internal = SMTP_INTERNAL_WRITE;
		return SMTP_POLL;
	} else if (res < linea_size) {
		/* Enviamos menos de lo que deberíamos o error, TODO: revisar este caso */
		smtp_session_close_and_clean (session);
		
		return SMTP_ERROR;
	}
	/* Envio completo */
	session->internal = SMTP_INTERNAL_READ;
	
	return smtp_session_continue (session, SMTP_WANT_POLL_IN);
}

int _smtp_session_internal_pass_code (SMTPSession *session) {
	session->estado = SMTP_STATE_READY;
	session->internal = SMTP_INTERNAL_FREE;
	if (session->last_code >= 200 && session->last_code < 300) {
		/* FIXME: activar la bandera de logueado */
		return SMTP_SUCCESS;
	}
	return SMTP_ERROR;
}

int _smtp_session_internal_send_cmd (SMTPSession *session) {
	int res, linea_size;
	
	linea_size = strlen (session->last_cmd);
	res = write (session->sock, session->last_cmd, linea_size);
	
	if (res < 0 && errno == EAGAIN) {
		session->poll = SMTP_WANT_POLL_OUT;
		session->internal = SMTP_INTERNAL_WRITE;
		return SMTP_POLL;
	} else if (res < linea_size) {
		/* Enviamos menos de lo que deberíamos o error, TODO: revisar este caso */
		smtp_session_close_and_clean (session);
		
		return SMTP_ERROR;
	}
	/* Envio completo */
	session->internal = SMTP_INTERNAL_READ;
	
	return smtp_session_continue (session, SMTP_WANT_POLL_IN);
}

int _smtp_session_internal_cmd_code (SMTPSession *session) {
	session->estado = SMTP_STATE_READY;
	session->internal = SMTP_INTERNAL_FREE;
	
	return SMTP_SUCCESS;
}

int _smtp_session_internal_data_write_finish (SMTPSession *session) {
	session->estado = SMTP_STATE_READY;
	session->internal = SMTP_INTERNAL_FREE;
	if (session->last_code >= 200 && session->last_code < 300) {
		return SMTP_SUCCESS;
	}
	return SMTP_ERROR;
}

int _smtp_session_internal_data_write (SMTPSession *session) {
	int res;
	char buffer[8192];
	size_t consumed;
	long pos;
	
	if (session->data_stream == NULL) {
		/* Si no hay flujo, nada que mandar. Cerrar canal de datos */
		
		/* Pasar a leer el 200 del comando */
		session->internal = SMTP_INTERNAL_READ;
		
		return smtp_session_continue (session, SMTP_WANT_POLL_IN);
	}

smtp_data_write_again:
	/* Primero hacer una lectura del flujo */
	pos = ftell (session->data_stream);
	consumed = fread (buffer, 1, sizeof (buffer), session->data_stream);
	if (consumed == 0) {
		//if (feof (session->data_stream) || ferror (session->data_stream)) {
			/* Fin de archivo, continuar con los comandos */
			session->data_stream = NULL;
			
			/* Pasar a leer el 200 del comando */
			session->internal = SMTP_INTERNAL_READ;
			
			return smtp_session_continue (session, SMTP_WANT_POLL_IN);
		//}
	}
	
	/* Tratar de escribir la cantidad leida */
	res = send (session->sock, buffer, consumed, 0);
	if (res > 0) {
		/* Reajustar el flujo de entrada */
		pos = pos + res;
		
		fseek (session->data_stream, pos, SEEK_SET);
		
		/* Reintentar enviar datos */
		goto smtp_data_write_again;
	} else if (res <= 0) {
		if (errno == EAGAIN || errno == EINTR) {
			/* Bloquearía, reintentar después */
			fseek (session->data_stream, pos, SEEK_SET);
			
			session->poll = SMTP_WANT_POLL_OUT;
			return SMTP_POLL;
		}
		
		/* En otro error, se cerró la conexión de datos */
		session->data_stream = NULL;
		session->internal = SMTP_INTERNAL_READ;
		
		return smtp_session_continue (session, SMTP_WANT_POLL_IN);
	}
}

int _smtp_session_internal_send_close (SMTPSession *session) {
	char *linea;
	int res, linea_size;
	
	linea = "QUIT\r\n";
	linea_size = strlen (linea);
	res = write (session->sock, linea, linea_size);
	
	if (res < 0 && errno == EAGAIN) {
		session->poll = SMTP_WANT_POLL_OUT;
		session->internal = SMTP_INTERNAL_WRITE;
		return SMTP_POLL;
	} else if (res < linea_size) {
		/* Enviamos menos de lo que deberíamos o error, TODO: revisar este caso */
		smtp_session_close_and_clean (session);
		
		return SMTP_ERROR;
	}
	/* Envio completo */
	session->internal = SMTP_INTERNAL_READ;
	
	return smtp_session_continue (session, SMTP_WANT_POLL_IN);
}

int _smtp_session_internal_close_code (SMTPSession *session) {
	if (session->last_code >= 200 && session->last_code < 300) {
		session->internal = SMTP_INTERNAL_FREE;
		
		/* Intencionalmente no cambiamos de estado para que nos llegue el cierre */
		return smtp_session_continue (session, SMTP_WANT_POLL_IN);
	} else {
		smtp_session_close_and_clean (session);
	}
	
	return SMTP_SUCCESS;
}

int _smtp_session_internal_send_data (SMTPSession *session) {
	char *linea;
	int res, linea_size;
	
	linea = "DATA\r\n";
	linea_size = strlen (linea);
	res = write (session->sock, linea, linea_size);
	
	if (res < 0 && errno == EAGAIN) {
		session->poll = SMTP_WANT_POLL_OUT;
		session->internal = SMTP_INTERNAL_WRITE;
		return SMTP_POLL;
	} else if (res < linea_size) {
		/* Enviamos menos de lo que deberíamos o error, TODO: revisar este caso */
		smtp_session_close_and_clean (session);
		
		return SMTP_ERROR;
	}
	/* Envio completo */
	session->internal = SMTP_INTERNAL_READ;
	
	return smtp_session_continue (session, SMTP_WANT_POLL_IN);
}

int _smtp_session_internal_data_code (SMTPSession *session) {
	int c;
	int res;
	
	if (session->last_code >= 300 && session->last_code < 400) {
		/* Avanzamos de estado */
		session->internal = SMTP_INTERNAL_WRITE;
		session->estado = SMTP_STATE_SENDING_DATA;
		
		return smtp_session_continue (session, SMTP_WANT_POLL_OUT);
	} else {
		/* En el caso de un código de error, regresar */
		session->internal = SMTP_INTERNAL_FREE;
		session->estado = SMTP_STATE_READY;
		
		return SMTP_ERROR;
	}
}

int smtp_session_read_bytes (SMTPSession *session) {
	int res;
	if (session->read_count >= sizeof (session->read_buffer)) {
		return SMTP_ERROR;
	}
	
	/* Leer tantos bytes como tenga el buffer */
	res = read (session->sock, &session->read_buffer[session->read_count], sizeof (session->read_buffer) - session->read_count);
	
	if (res == 0) {
		session->ceof = 1;
	} else if (res < 0) {
		if (errno == EINTR || errno == EAGAIN) {
			return SMTP_POLL;
		}
		return SMTP_ERROR;
	} else {
		session->read_count += res;
	}
	
	return SMTP_SUCCESS;
}

int smtp_session_readline (SMTPSession *session, char *buffer, int max) {
	char *end;
	int line_size = 0;
	
	if (session->read_count > 0) {
		/* Ya había otros bytes en el buffer, procesarlos para buscar el salto de linea aquí */
		line_size = (max >= session->read_count) ? session->read_count : max - 1;
		
		end = memccpy (buffer, session->read_buffer, '\n', line_size);
		
		if (end != NULL) {
			line_size = end - buffer;
		}
		
		buffer[line_size] = 0;
		
		/* Consumir x bytes de la entrada */
		if (line_size < session->read_count) {
			memmove (session->read_buffer, &session->read_buffer[line_size], session->read_count - line_size);
			
			session->read_count -= line_size;
		} else {
			session->read_buffer[0] = 0;
			session->read_count = 0;
		}
		
		if (end != NULL) {
			
			if (buffer[line_size - 2] == '\r') {
				/* Quitar el \r */
				buffer[line_size - 2] = '\n';
				buffer[line_size - 1] = 0;
				line_size--;
			}
			
			return line_size;
		}
	}
	
	return SMTP_POLL;	
}

int smtp_session_get_state (SMTPSession *session) {
	return session->estado;
}

int smtp_session_auth_login (SMTPSession *session, const char *user, const char *passwd) {
	if (session->estado != SMTP_STATE_READY || session->internal != SMTP_INTERNAL_FREE) {
		return SMTP_ERROR;
	}
	
	/* Copiar el usuario y contraseña */
	session->user = strdup (user);
	if (passwd != NULL) {
		session->passwd = strdup (passwd);
	} else {
		session->passwd = NULL;
	}
	
	session->estado = SMTP_STATE_AUTH_LOGIN;
	session->internal = SMTP_INTERNAL_WRITE;
	return smtp_session_continue (session, SMTP_WANT_POLL_OUT);
}

int smtp_session_send_command (SMTPSession *session, char *command) {
	int len;
	
	/* TODO: Revisar aquí otros estados */
	if (session->estado != SMTP_STATE_READY || session->internal != SMTP_INTERNAL_FREE) {
		return SMTP_ERROR;
	}
	
	/* Sanitizar el comando y asegurarnos que contenga el \r\n */
	len = strlen (command);
	if (command[len - 2] == '\r' && command[len - 1] == '\n') {
		if (len + 1 > sizeof (session->last_cmd)) {
			return SMTP_ERROR;
		}
		strncpy (session->last_cmd, command, sizeof (session->last_cmd));
	} else if (command[len - 1] == '\r') {
		if (len + 2 > sizeof (session->last_cmd)) {
			return SMTP_ERROR;
		}
		strncpy (session->last_cmd, command, sizeof (session->last_cmd));
		session->last_cmd[len] = '\n';
		session->last_cmd[len + 1] = 0;
	} else {
		if (len + 3 > sizeof (session->last_cmd)) {
			return SMTP_ERROR;
		}
		strncpy (session->last_cmd, command, sizeof (session->last_cmd));
		session->last_cmd[len] = '\r';
		session->last_cmd[len + 1] = '\n';
		session->last_cmd[len + 2] = 0;
	}
	
	session->estado = SMTP_STATE_CMD;
	session->internal = SMTP_INTERNAL_WRITE;
	
	return smtp_session_continue (session, SMTP_WANT_POLL_OUT);
}

int smtp_session_send_close (SMTPSession *session) {
	char *command;
	int len;
	
	if (session->estado != SMTP_STATE_READY || session->internal != SMTP_INTERNAL_FREE) {
		return SMTP_ERROR;
	}
	
	session->estado = SMTP_STATE_CLOSING;
	session->internal = SMTP_INTERNAL_WRITE;
	
	return smtp_session_continue (session, SMTP_WANT_POLL_OUT);
}

int smtp_session_send_data (SMTPSession *session, FILE *stream) {
	char *command;
	int len;
	
	if (session->estado != SMTP_STATE_READY || session->internal != SMTP_INTERNAL_FREE) {
		return SMTP_ERROR;
	}
	
	session->estado = SMTP_STATE_DATA;
	session->internal = SMTP_INTERNAL_WRITE;
	session->data_stream = stream;
	
	return smtp_session_continue (session, SMTP_WANT_POLL_OUT);
}

int smtp_session_get_socket (SMTPSession *session) {
	return session->sock;
}

int smtp_session_get_last_code (SMTPSession *session) {
	return session->last_code;
}

char * smtp_session_get_last_message (SMTPSession *session) {
	return session->last_message;
}

void smtp_session_multiline_func (SMTPSession *session, SMTPLineFunc func, void *data) {
	session->line_func = func;
	session->line_func_data = data;
}

int smtp_session_get_poll (SMTPSession *session) {
	return session->poll;
}

